
type expr =
  | Var of int
  | Cst of int
  | Add of expr * expr
  | Mul of expr * expr
  | Sub of expr * expr

type in_aexpr =
  | AVar of int
  | ACst of int
  | AAdd of aexpr * aexpr
  | AMul of aexpr * aexpr
  | ASub of aexpr * aexpr
and aexpr = { pos : int ; expr : in_aexpr }

(** Annotation de l'AST
  * selon un parcours en profondeur, de gauche à droite, post-fixe. *)
let rec annot cur expr =
  match expr with
    | Var i -> { pos = cur ; expr = AVar i }, cur+1
    | Cst i -> { pos = cur ; expr = ACst i }, cur+1
    | Add (e1,e2) ->
        let a1,cur = annot cur e1 in
        let a2,cur = annot cur e2 in
          { pos = cur ; expr = AAdd (a1,a2) }, cur+1
    | Mul (e1,e2) ->
        let a1,cur = annot cur e1 in
        let a2,cur = annot cur e2 in
          { pos = cur ; expr = AMul (a1,a2) }, cur+1
    | Sub (e1,e2) ->
        let a1,cur = annot cur e1 in
        let a2,cur = annot cur e2 in
          { pos = cur ; expr = ASub (a1,a2) }, cur+1

(** Offset d'une aexpr dans la pile. *)
let pos aexpr = 4*aexpr.pos

(** Compilation d'une aexpr: génération de l'assembleur qui calcule
  * la valeur de l'expression et la stocke à la position correspondante
  * dans la pile. *)
let rec compile_aexpr aexpr =
  match aexpr.expr with
    | AVar i ->
        Printf.printf "  movq %d(%%rbp), %%rdi\n" (8*i) ;
        Printf.printf "  call atoi\n" ;
        Printf.printf "  movl %%eax, %d(%%rsp)\n" (pos aexpr)
    | ACst i ->
        Printf.printf "  movl $%d, %d(%%rsp)\n" i (pos aexpr)
    | AAdd (a1,a2) ->
        compile_aexpr a1 ;
        compile_aexpr a2 ;
        Printf.printf "  movl %d(%%rsp), %%eax\n" (pos a1) ;
        Printf.printf "  addl %d(%%rsp), %%eax\n" (pos a2) ;
        Printf.printf "  movl %%eax, %d(%%rsp)\n" (pos aexpr)
    | AMul (a1,a2) ->
        compile_aexpr a1 ;
        compile_aexpr a2 ;
        Printf.printf "  movl %d(%%rsp), %%eax\n" (pos a1) ;
        Printf.printf "  imul %d(%%rsp), %%eax\n" (pos a2) ;
        Printf.printf "  movl %%eax, %d(%%rsp)\n" (pos aexpr)
    | ASub (a1,a2) ->
        compile_aexpr a1 ;
        compile_aexpr a2 ;
        Printf.printf "  movl %d(%%rsp), %%eax\n" (pos a1) ;
        Printf.printf "  subl %d(%%rsp), %%eax\n" (pos a2) ;
        Printf.printf "  movl %%eax, %d(%%rsp)\n" (pos aexpr)

let compile expr =
  let aexpr,cur = annot 0 expr in
    Printf.printf ".text\nmain:\n.global main\n" ;
    (* On stocke RDI (2ème argument de main = argv) dans RBP.
     * C'est nécessaire car RDI est caller-saved et on va appeler la fonction
     * atoi. *)
    Printf.printf "  movq %%rsi, %%rbp\n" ;
    (* Allocation de l'espace pour stocker cur entiers 32 bits. *)
    Printf.printf "  addq $%d, %%rsp\n" (-4*cur) ;
    compile_aexpr aexpr ;
    (* Affichage du résultat. *)
    Printf.printf
      "  movl %d(%%rsp), %%edi\ncall putint\n"
      (pos aexpr) ;
    (* On termine proprement. *)
    Printf.printf "  addq $%d, %%rsp\nmovl $0, %%eax\nret\n" (4*cur)

let () =
  (* Devrait afficher 0xbeef avec argv[1]="16" *)
  compile (Sub (Add (Cst 0xff, Mul (Cst 0xbe, Cst 0x100)), Var 1))
